# Generated by Django 4.2.5 on 2023-10-09 15:35

import datetime
import uuid

import django.db.models.deletion
from django.db import migrations, models

import django_napse.utils.constants
import django_napse.utils.findable_class


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Architecture",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            bases=(models.Model, django_napse.utils.findable_class.FindableClass),
        ),
        migrations.CreateModel(
            name="Bot",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                ("name", models.CharField(default="Bot", max_length=100)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("active", models.BooleanField(default=True)),
            ],
        ),
        migrations.CreateModel(
            name="BotConfig",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                ("immutable", models.BooleanField(default=False)),
            ],
            bases=(models.Model, django_napse.utils.findable_class.FindableClass),
        ),
        migrations.CreateModel(
            name="Cluster",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("share", models.FloatField()),
                ("breakpoint", models.FloatField()),
                ("autoscale", models.BooleanField()),
            ],
        ),
        migrations.CreateModel(
            name="Connection",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "bot",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="connections",
                        to="django_napse_core.bot",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Controller",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("pair", models.CharField(max_length=10)),
                ("base", models.CharField(max_length=10)),
                ("quote", models.CharField(max_length=10)),
                ("interval", models.CharField(max_length=10)),
                ("min_notional", models.FloatField(null=True)),
                ("min_trade", models.FloatField(null=True)),
                ("lot_size", models.IntegerField(null=True)),
                ("price", models.FloatField(null=True)),
                ("last_price_update", models.DateTimeField(null=True)),
                ("last_settings_update", models.DateTimeField(null=True)),
            ],
        ),
        migrations.CreateModel(
            name="Exchange",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=200, unique=True)),
                ("description", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="ExchangeAccount",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                ("name", models.CharField(max_length=200, unique=True)),
                ("testing", models.BooleanField(default=True)),
                ("description", models.TextField()),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "exchange",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="accounts",
                        to="django_napse_core.exchange",
                    ),
                ),
            ],
            bases=(models.Model, django_napse.utils.findable_class.FindableClass),
        ),
        migrations.CreateModel(
            name="Modification",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        default=django_napse.utils.constants.MODIFICATION_STATUS["PENDING"],
                        max_length=15,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("ignore_failed_order", models.BooleanField(default=False)),
                ("key", models.CharField(max_length=100)),
                ("value", models.CharField(max_length=100)),
                ("target_type", models.CharField(max_length=100)),
            ],
            bases=(models.Model, django_napse.utils.findable_class.FindableClass),
        ),
        migrations.CreateModel(
            name="Plugin",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("category", models.CharField(max_length=255)),
            ],
            bases=(models.Model, django_napse.utils.findable_class.FindableClass),
        ),
        migrations.CreateModel(
            name="Strategy",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "architecture",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="strategy",
                        to="django_napse_core.architecture",
                    ),
                ),
                (
                    "config",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="strategy",
                        to="django_napse_core.botconfig",
                    ),
                ),
            ],
            bases=(models.Model, django_napse.utils.findable_class.FindableClass),
        ),
        migrations.CreateModel(
            name="Wallet",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(default="Wallet", max_length=255)),
                ("locked", models.BooleanField(default=False)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
            ],
            bases=(models.Model, django_napse.utils.findable_class.FindableClass),
        ),
        migrations.CreateModel(
            name="ArchitectureModification",
            fields=[
                (
                    "modification_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.modification",
                    ),
                ),
            ],
            bases=("django_napse_core.modification",),
        ),
        migrations.CreateModel(
            name="DCABotConfig",
            fields=[
                (
                    "botconfig_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.botconfig",
                    ),
                ),
                ("setting_timeframe", models.DurationField()),
            ],
            bases=("django_napse_core.botconfig",),
        ),
        migrations.CreateModel(
            name="DCAStrategy",
            fields=[
                (
                    "strategy_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.strategy",
                    ),
                ),
                ("variable_last_buy_date", models.DateTimeField(blank=True, null=True)),
            ],
            bases=("django_napse_core.strategy",),
        ),
        migrations.CreateModel(
            name="EmptyBotConfig",
            fields=[
                (
                    "botconfig_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.botconfig",
                    ),
                ),
                ("setting_empty", models.BooleanField()),
            ],
            bases=("django_napse_core.botconfig",),
        ),
        migrations.CreateModel(
            name="EmptyStrategy",
            fields=[
                (
                    "strategy_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.strategy",
                    ),
                ),
            ],
            bases=("django_napse_core.strategy",),
        ),
        migrations.CreateModel(
            name="LBOPlugin",
            fields=[
                (
                    "plugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.plugin",
                    ),
                ),
            ],
            bases=("django_napse_core.plugin",),
        ),
        migrations.CreateModel(
            name="MBPPlugin",
            fields=[
                (
                    "plugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.plugin",
                    ),
                ),
            ],
            bases=("django_napse_core.plugin",),
        ),
        migrations.CreateModel(
            name="SBVPlugin",
            fields=[
                (
                    "plugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.plugin",
                    ),
                ),
            ],
            bases=("django_napse_core.plugin",),
        ),
        migrations.CreateModel(
            name="StrategyModification",
            fields=[
                (
                    "modification_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.modification",
                    ),
                ),
            ],
            bases=("django_napse_core.modification",),
        ),
        migrations.CreateModel(
            name="TurboDCABotConfig",
            fields=[
                (
                    "botconfig_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.botconfig",
                    ),
                ),
                ("setting_timeframe", models.DurationField()),
                ("setting_percentage", models.FloatField()),
            ],
            bases=("django_napse_core.botconfig",),
        ),
        migrations.CreateModel(
            name="TurboDCAStrategy",
            fields=[
                (
                    "strategy_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.strategy",
                    ),
                ),
                ("variable_last_buy_date", models.DateTimeField(blank=True, null=True)),
            ],
            bases=("django_napse_core.strategy",),
        ),
        migrations.CreateModel(
            name="Transaction",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("amount", models.FloatField()),
                ("ticker", models.CharField(max_length=10)),
                (
                    "transaction_type",
                    models.CharField(default="TRANSFER", max_length=20),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "from_wallet",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="transactions_from",
                        to="django_napse_core.wallet",
                    ),
                ),
                (
                    "to_wallet",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="transactions_to",
                        to="django_napse_core.wallet",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="plugin",
            name="strategy",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="plugins",
                to="django_napse_core.strategy",
            ),
        ),
        migrations.CreateModel(
            name="OrderBatch",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        default=django_napse.utils.constants.ORDER_STATUS["PENDING"],
                        max_length=15,
                    ),
                ),
                (
                    "controller",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="order_batches",
                        to="django_napse_core.controller",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Order",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("price", models.FloatField()),
                ("pair", models.CharField(max_length=10)),
                ("side", models.CharField(max_length=10)),
                ("completed", models.BooleanField(default=False)),
                ("asked_for_amount", models.FloatField()),
                ("asked_for_ticker", models.CharField(max_length=10)),
                ("debited_amount", models.FloatField(default=0)),
                ("batch_share", models.FloatField(default=0)),
                ("exit_base_amount", models.FloatField(default=0)),
                ("exit_quote_amount", models.FloatField(default=0)),
                ("fees", models.FloatField(default=0)),
                ("fee_ticker", models.CharField(blank=True, max_length=10)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "batch",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="orders",
                        to="django_napse_core.orderbatch",
                    ),
                ),
                (
                    "connection",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="orders",
                        to="django_napse_core.connection",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="NapseSpace",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=200)),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                ("description", models.TextField()),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "exchange_account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="spaces",
                        to="django_napse_core.exchangeaccount",
                    ),
                ),
            ],
            options={
                "unique_together": {("name", "exchange_account")},
            },
        ),
        migrations.AddField(
            model_name="modification",
            name="order",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="modifications",
                to="django_napse_core.order",
            ),
        ),
        migrations.CreateModel(
            name="Link",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("importance", models.FloatField()),
                (
                    "bot",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="link",
                        to="django_napse_core.bot",
                    ),
                ),
                (
                    "cluster",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links",
                        to="django_napse_core.cluster",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Fleet",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                ("name", models.CharField(default="Fleet", max_length=100)),
                ("running", models.BooleanField(default=False)),
                ("setup_finished", models.BooleanField(default=False)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "exchange_account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_napse_core.exchangeaccount",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Debit",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("amount", models.FloatField()),
                ("ticker", models.CharField(max_length=10)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "wallet",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="debits",
                        to="django_napse_core.wallet",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Credit",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("amount", models.FloatField()),
                ("ticker", models.CharField(max_length=10)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "wallet",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="credits",
                        to="django_napse_core.wallet",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="controller",
            name="exchange_account",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="controller",
                to="django_napse_core.exchangeaccount",
            ),
        ),
        migrations.AddField(
            model_name="connection",
            name="owner",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="connections",
                to="django_napse_core.wallet",
            ),
        ),
        migrations.AddField(
            model_name="cluster",
            name="fleet",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="clusters",
                to="django_napse_core.fleet",
            ),
        ),
        migrations.AddField(
            model_name="cluster",
            name="template_bot",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="cluster",
                to="django_napse_core.bot",
            ),
        ),
        migrations.AddField(
            model_name="botconfig",
            name="space",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="bot_configs",
                to="django_napse_core.napsespace",
            ),
        ),
        migrations.AddField(
            model_name="bot",
            name="strategy",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="bot",
                to="django_napse_core.strategy",
            ),
        ),
        migrations.CreateModel(
            name="SpaceWallet",
            fields=[
                (
                    "wallet_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.wallet",
                    ),
                ),
                (
                    "owner",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="wallet",
                        to="django_napse_core.napsespace",
                    ),
                ),
            ],
            bases=("django_napse_core.wallet",),
        ),
        migrations.CreateModel(
            name="SpaceSimulationWallet",
            fields=[
                (
                    "wallet_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.wallet",
                    ),
                ),
                (
                    "owner",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="simulation_wallet",
                        to="django_napse_core.napsespace",
                    ),
                ),
            ],
            bases=("django_napse_core.wallet",),
        ),
        migrations.CreateModel(
            name="SinglePairArchitecture",
            fields=[
                (
                    "architecture_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.architecture",
                    ),
                ),
                (
                    "variable_last_candle_date",
                    models.DateTimeField(
                        default=datetime.datetime(
                            1970,
                            1,
                            1,
                            0,
                            0,
                            tzinfo=datetime.timezone.utc,
                        ),
                    ),
                ),
                (
                    "controller",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="single_pair_architectures",
                        to="django_napse_core.controller",
                    ),
                ),
            ],
            bases=("django_napse_core.architecture",),
        ),
        migrations.CreateModel(
            name="OrderWallet",
            fields=[
                (
                    "wallet_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.wallet",
                    ),
                ),
                (
                    "owner",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="wallet",
                        to="django_napse_core.order",
                    ),
                ),
            ],
            bases=("django_napse_core.wallet",),
        ),
        migrations.CreateModel(
            name="Currency",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("mbp", models.FloatField()),
                ("ticker", models.CharField(max_length=10)),
                ("amount", models.FloatField(default=0)),
                (
                    "wallet",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="currencies",
                        to="django_napse_core.wallet",
                    ),
                ),
            ],
            options={
                "unique_together": {("wallet", "ticker")},
            },
        ),
        migrations.AlterUniqueTogether(
            name="controller",
            unique_together={("pair", "interval", "exchange_account")},
        ),
        migrations.CreateModel(
            name="ConnectionWallet",
            fields=[
                (
                    "wallet_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.wallet",
                    ),
                ),
                (
                    "owner",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="wallet",
                        to="django_napse_core.connection",
                    ),
                ),
            ],
            bases=("django_napse_core.wallet",),
        ),
        migrations.CreateModel(
            name="ConnectionSpecificArgs",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("key", models.CharField(max_length=100)),
                ("value", models.CharField(default="None", max_length=100)),
                ("target_type", models.CharField(default="None", max_length=100)),
                (
                    "connection",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="specific_args",
                        to="django_napse_core.connection",
                    ),
                ),
            ],
            options={
                "unique_together": {("connection", "key")},
            },
        ),
        migrations.CreateModel(
            name="ConnectionModification",
            fields=[
                (
                    "modification_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.modification",
                    ),
                ),
                (
                    "connection_specific_arg",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="modifications",
                        to="django_napse_core.connectionspecificargs",
                    ),
                ),
            ],
            bases=("django_napse_core.modification",),
        ),
        migrations.AlterUniqueTogether(
            name="connection",
            unique_together={("owner", "bot")},
        ),
        migrations.CreateModel(
            name="BinanceAccount",
            fields=[
                (
                    "exchangeaccount_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.exchangeaccount",
                    ),
                ),
                ("public_key", models.CharField(max_length=200)),
                ("private_key", models.CharField(max_length=200)),
            ],
            options={
                "unique_together": {("public_key", "private_key")},
            },
            bases=("django_napse_core.exchangeaccount",),
        ),
    ]



# Generated by Django 4.2.7 on 2024-02-29 22:33

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("django_napse_auth", "0005_alter_keypermission_space"),
        ("django_napse_simulations", "0001_initial"),
        ("django_napse_core", "0009_rename_exit_base_amount_order_exit_amount_base_and_more"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="NapseSpace",
            new_name="Space",
        ),
    ]



from django.db import models

from django_napse.utils.errors import BotConfigError


class BotConfigManager(models.Manager):
    def create(
        self,
        space,
        immutable: bool = False,
        settings=None,
    ):
        settings = settings or {}
        for setting in self.model._meta.get_fields():
            if setting.name.startswith("setting_"):
                try:
                    settings[setting.name[8:]]
                except KeyError as e:
                    error_msg = f"Missing setting: {setting.name[8:]}"
                    raise BotConfigError.MissingSettingError(error_msg) from e
        if not immutable:
            try:
                self.model.objects.get(
                    space=space,
                    immutable=immutable,
                    **{f"setting_{setting_name}": setting_value for setting_name, setting_value in settings.items()},
                )
            except self.model.DoesNotExist:
                pass
            else:
                error_msg = "This BotConfig already exists in this space."
                raise BotConfigError.DuplicateBotConfig(error_msg)
        config = self.model(
            space=space,
            immutable=immutable,
            **{f"setting_{setting_name}": setting_value for setting_name, setting_value in settings.items()},
        )
        config.save()
        return config



from .architecture_serializer import ArchitectureSerializer
from .bot_serializers import BotDetailSerializer, BotSerializer
from .config_serializer import ConfigSerializer
from .plugin_serializer import PluginSerializer
from .strategy_serializer import StrategySerializer



import json
import logging
from pathlib import Path

import environ
from django.conf import settings

from django_napse.utils.constants import EXCHANGES
from django_napse.utils.errors import NapseError

logger = logging.getLogger("django_napse")


class DjangoNapseSettings:
    """Main django_napse settings."""

    @property
    def NAPSE_EXCHANGE_CONFIGS(self) -> dict[str, dict[str, any]]:  # noqa: D102
        return getattr(
            settings,
            "NAPSE_EXCHANGE_CONFIGS",
            {
                "BINANCE": {
                    "description": "Binance exchange. More info: https://www.binance.com/en",
                },
            },
        )

    @property
    def NAPSE_IS_IN_PIPELINE(self) -> bool:  # noqa: D102
        return environ.Env().bool("NAPSE_IS_IN_PIPELINE", default=False)

    @property
    def NAPSE_SECRETS_FILE_PATH(self) -> Path:  # noqa: D102
        if isinstance(settings.NAPSE_SECRETS_FILE_PATH, str):
            return Path(settings.NAPSE_SECRETS_FILE_PATH).absolute()

        if isinstance(settings.NAPSE_SECRETS_FILE_PATH, Path):
            return settings.NAPSE_SECRETS_FILE_PATH

        error_msg: str = "NAPSE_SECRETS_FILE_PATH must be a string or a Path object."
        raise ValueError(error_msg)

    @property
    def NAPSE_ENV_FILE_PATH(self) -> Path:  # noqa: D102
        if isinstance(settings.NAPSE_ENV_FILE_PATH, str):
            return Path(settings.NAPSE_ENV_FILE_PATH).absolute()

        if isinstance(settings.NAPSE_ENV_FILE_PATH, Path):
            return settings.NAPSE_ENV_FILE_PATH

        error_msg: str = "NAPSE_ENV_FILE_PATH must be a string or a Path object."
        raise ValueError(error_msg)

    @property
    def NAPSE_EXCHANGES_TO_TEST(self) -> list[str]:  # noqa: D102
        return getattr(settings, "NAPSE_EXCHANGES_TO_TEST", ["BINANCE"])

    @property
    def NAPSE_MASTER_KEY(self) -> str:  # noqa: D102
        return getattr(settings, "NAPSE_MASTER_KEY", "insecure_master_key")


napse_settings = DjangoNapseSettings()

if settings.configured:
    for app in [
        "rest_framework",
        "rest_framework_api_key",
        "django_celery_beat",
        "corsheaders",
    ]:
        if app not in settings.INSTALLED_APPS:
            warning = f"{app} not found in settings.INSTALLED_APPS. Please add it to your settings file."
            logger.warning(warning)
            settings.INSTALLED_APPS += (app,)

    if "REST_FRAMEWORK" not in settings.__dir__():
        warning = "REST_FRAMEWORK not found in settings. Please add it to your settings file."
        logger.warning(warning)
        settings.REST_FRAMEWORK = {
            "DEFAULT_PERMISSION_CLASSES": ["django_napse.api.custom_permissions.HasAdminPermission"],
        }
    for permission in ["django_napse.api.custom_permissions.HasAdminPermission"]:
        if permission not in settings.REST_FRAMEWORK["DEFAULT_PERMISSION_CLASSES"]:
            warning = f"{permission} not found in settings. Please add it to your settings file."
            logger.warning(warning)

    if "NAPSE_SECRETS_FILE_PATH" not in settings.__dir__():
        warning = "NAPSE_SECRETS_FILE_PATH not found in settings. Please add it to your settings file."
        logger.warning(warning)
    if not napse_settings.NAPSE_SECRETS_FILE_PATH.exists():
        logger.warning(
            "WARNING: No secrets file found at %s. Creating one now.",
            napse_settings.NAPSE_SECRETS_FILE_PATH,
        )
        secrets_file = napse_settings.NAPSE_SECRETS_FILE_PATH.open("w")
        json.dump({"Exchange Accounts": {}}, secrets_file)
        secrets_file.close()

    if "NAPSE_ENV_FILE_PATH" not in settings.__dir__():
        warning = "NAPSE_ENV_FILE_PATH not found in settings. Please add it to your settings file."
        logger.warning(warning)

    if not napse_settings.NAPSE_ENV_FILE_PATH.exists():
        logger.warning(
            "WARNING: No env file found at %s. Creating one now.",
            napse_settings.NAPSE_ENV_FILE_PATH,
        )
        env_file = napse_settings.NAPSE_ENV_FILE_PATH.open("w")
        env_file.write("")
        env_file.close()

    if {*list(napse_settings.NAPSE_EXCHANGE_CONFIGS.keys())} != set(EXCHANGES):
        error_msg = "NAPSE_EXCHANGE_CONFIGS does not match the list of exchanges. Can't start the server."
        raise NapseError.SettingsError(error_msg)



from django.db import models

from django_napse.core.models.transactions.managers import DebitManager


class Debit(models.Model):
    """A Debit is a transaction that forcefully removes money from a wallet."""

    wallet = models.ForeignKey("Wallet", on_delete=models.CASCADE, related_name="debits")
    amount = models.FloatField()
    ticker = models.CharField(max_length=10)
    created_at = models.DateTimeField(auto_now_add=True)

    objects = DebitManager()

    def __str__(self) -> str:  # pragma: no cover
        return f"DEBIT: {self.pk})"

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return a string with the model information.

        Args:
            beacon (str, optional): The prefix for each line. Defaults to "".
            verbose (bool, optional): Whether to print the string. Defaults to True.

        Returns:
            str: The string with the history information.
        """
        string = ""
        string += f"{beacon}Debit {self.pk=}\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.amount=}\n"
        string += f"{beacon}\t{self.ticker=}\n"
        string += f"{beacon}Debited Wallet:\n"
        wall_str = self.wallet.info(verbose=False, beacon=beacon + "\t")
        string += f"{wall_str}\n"

        if verbose:  # pragma: no cover
            print(string)
        return string



from django.db import models

from django_napse.core.models.bots.architectures.single_pair import SinglePairArchitecture
from django_napse.core.models.bots.implementations.dca.config import DCABotConfig
from django_napse.core.models.bots.strategy import Strategy
from django_napse.utils.constants import SIDES


class DCAStrategy(Strategy):
    variable_last_buy_date = models.DateTimeField(null=True, blank=True)

    def __str__(self) -> str:
        return f"DCA BOT STRATEGY: {self.pk=}"

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}Strategy ({self.pk=}):\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.config=}\n"
        string += f"{beacon}\t{self.architecture=}\n"
        if verbose:  # pragma: no cover
            print(string)
        return string

    @classmethod
    def config_class(cls):
        return DCABotConfig

    @classmethod
    def architecture_class(cls):
        return SinglePairArchitecture

    def give_order(self, data: dict) -> list[dict]:
        controller = data["controllers"]["main"]
        if (
            self.variable_last_buy_date is None
            or data["candles"][controller]["current"]["open_time"] - self.variable_last_buy_date >= data["config"]["timeframe"]
        ):
            return [
                {
                    "controller": controller,
                    "ArchitectureModifications": [],
                    "StrategyModifications": [
                        {
                            "key": "last_buy_date",
                            "value": str(data["candles"][controller]["current"]["open_time"]),
                            "target_type": "datetime",
                            "ignore_failed_order": False,
                        },
                    ],
                    "ConnectionModifications": [],
                    "connection": data["connection"],
                    "asked_for_amount": 20,
                    "asked_for_ticker": controller.quote,
                    "pair": controller.pair,
                    "price": data["candles"][controller]["latest"]["close"],
                    "side": SIDES.BUY,
                },
            ]
        return [
            {
                "controller": controller,
                "ArchitectureModifications": [],
                "StrategyModifications": [],
                "ConnectionModifications": [],
                "connection": data["connection"],
                "asked_for_amount": 0,
                "asked_for_ticker": controller.quote,
                "pair": controller.pair,
                "price": data["candles"][controller]["latest"]["close"],
                "side": SIDES.KEEP,
            },
        ]



from django.db.models import QuerySet
from rest_framework import status
from rest_framework.response import Response
from rest_framework_api_key.permissions import HasAPIKey

from django_napse.api.bots.serializers.bot_serializers import BotDetailSerializer, BotSerializer
from django_napse.api.custom_permissions import HasSpace
from django_napse.api.custom_viewset import CustomViewSet
from django_napse.core.models import Bot, Space
from django_napse.utils.errors import APIError


class BotView(CustomViewSet):
    permission_classes = [HasAPIKey, HasSpace]
    serializer_class = BotSerializer

    def get_queryset(self) -> list[QuerySet[Bot]] | dict[str, QuerySet[Bot]]:
        """Return bot queryset.

        Can return
            Free bots across all available spaces
            Bots with connections without space containerization
            Bots with connections with a specific space containerization
            Bots with connections with space containerization

        Raises:
            ValueError: Not space_containers mode is only available for master key.
            ValueError: Space not found.
        """
        api_key = self.get_api_key(self.request)
        spaces = Space.objects.all() if api_key.is_master_key else [permission.space for permission in api_key.permissions.all()]

        # Free bots across all available spaces
        if self.request.query_params.get("free", False):
            # Exchange account containerization
            request_space = self.get_space(self.request)
            if request_space is None:
                raise APIError.MissingSpace()
            spaces = [space for space in spaces if space.exchange_account == request_space.exchange_account]
            # Cross space free bots
            return [bot for bot in Bot.objects.filter(strategy__config__space__in=spaces) if bot.is_free]

        # Bots with connections without space containerization
        if not self.request.query_params.get("space_containers", True):
            if not api_key.is_master_key:
                error_msg: str = "Not space_containers mode is only available for master key."
                raise ValueError(error_msg)
            return Bot.objects.exclude(connections__isnull=True)

        # Filter by specific space
        space_uuid = self.request.query_params.get("space", None)
        if space_uuid is not None:
            for space in spaces:
                if space.uuid == space_uuid:
                    spaces = [space]
                    break
            else:
                error_msg: str = "Space not found."
                raise ValueError(error_msg)

        # Space container mode
        bots_per_space: dict[str, QuerySet[Bot]] = {}
        for space in spaces:
            bots_per_space[space] = space.bots.exclude(connections__isnull=True)
        return bots_per_space

    def get_serializer_class(self, *args, **kwargs):
        actions: dict = {
            "list": BotSerializer,
            "retrieve": BotDetailSerializer,
        }
        result = actions.get(self.action)
        return result if result else super().get_serializer_class()

    def get_permissions(self):
        match self.action:
            case "list" | "create":
                return [HasAPIKey()]
            case _:
                return super().get_permissions()

    def get_object(self):
        uuid = self.kwargs.get("pk", None)
        if uuid is None:
            return super().get_object()
        return Bot.objects.get(uuid=uuid)

    def _get_boolean_query_param(self, param: str) -> bool | None:
        """Return None if a boolean cannot be found."""
        if isinstance(param, bool):
            return param

        if not isinstance(param, str):
            return None

        match param.lower():
            case "true" | "1":
                return True
            case "false" | "0":
                return False
            case _:
                return None

    def list(self, request):
        """Return a list of bots.

        Warning: space_containers can lead to undesirable behaviour.
        """
        try:
            queryset = self.get_queryset()
        except ValueError as error:
            return Response({"detail": str(error)}, status=status.HTTP_400_BAD_REQUEST)

        if isinstance(queryset, list):
            serializer = self.get_serializer(queryset, many=True)
            return Response(serializer.data, status=status.HTTP_200_OK)

        if isinstance(queryset, dict):
            serialized_bots: QuerySet[Bot] = []
            for space, query in queryset.items():
                serializer = self.get_serializer(query, many=True, space=space)
                if serializer.data != []:
                    serialized_bots += serializer.data
            return Response(serialized_bots, status=status.HTTP_200_OK)

        return Response(status=status.HTTP_400_BAD_REQUEST)

    def retrieve(self, request, pk=None):
        instance = self.get_object()
        space = self.get_space(request)
        serializer = self.get_serializer(instance, space=space)
        return Response(serializer.data, status=status.HTTP_200_OK)



from django_napse.core.models.histories.wallet import WalletHistory
from django_napse.core.models.wallets.currency import Currency
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.db.histories.test_wallet_history -v2 --keepdb --parallel
"""


class WalletHistoryTestCase:
    model = WalletHistory

    def simple_create(self):
        wallet = self.space.wallet
        Currency.objects.create(wallet=wallet, ticker="BTC", amount=1, mbp=20000)
        Currency.objects.create(wallet=wallet, ticker="ETH", amount=0, mbp=1000)
        return wallet.history

    def test_generate_data_points(self):
        history = self.simple_create()
        history.generate_data_point()


class WalletHistoryBINANCETestCase(WalletHistoryTestCase, ModelTestCase):
    exchange = "BINANCE"



from .api import *
from .bots import *
from .exchange import *
from .fleets import *
from .histories import *
from .key import *
from .napse import *
from .orders import *
from .sims import *
from .spaces import *
from .tasks import *
from .transactions import *
from .wallets import *



from __future__ import annotations

import uuid
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from django.db import models
from django.utils.timezone import get_default_timezone

from django_napse.core.models.accounts.managers import SpaceManager
from django_napse.core.models.bots.bot import Bot
from django_napse.core.models.bots.controller import Controller
from django_napse.core.models.fleets.fleet import Fleet
from django_napse.core.models.orders.order import Order
from django_napse.core.models.transactions.credit import Credit
from django_napse.core.models.transactions.debit import Debit
from django_napse.utils.constants import EXCHANGE_TICKERS, STABLECOINS
from django_napse.utils.errors import SpaceError
from django_napse.utils.errors.exchange import ExchangeError
from django_napse.utils.errors.wallets import WalletError

if TYPE_CHECKING:
    from django_napse.core.models.wallets.currency import Currency


class Space(models.Model):
    """Categorize and manage money.

    Attributes:
        name: Name of the space.
        uuid: Unique identifier of the space.
        description: Description of the space.
        exchange_account: Exchange account of the space.
        created_at: Date of creation of the space.


    Examples:
        Create a space:
        ```python
        import django_napse.core.models import Space, ExchangeAccount

        exchange_account: ExchangeAccount = ...
        space = Space.objects.create(
            name="Space",
            description="Space description",
            exchange_account=exchange_account,
        )
        ```
    """

    name = models.CharField(max_length=200)
    uuid = models.UUIDField(
        default=uuid.uuid4,
        editable=False,
        unique=True,
    )
    description = models.TextField()
    exchange_account = models.ForeignKey("ExchangeAccount", on_delete=models.CASCADE, related_name="spaces")
    created_at = models.DateTimeField(auto_now_add=True)

    objects = SpaceManager()

    class Meta:  # noqa: D106
        unique_together = ("name", "exchange_account")

    def __str__(self) -> str:
        return f"SPACE: {self.name}"

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return a string with the model information.

        Args:
            beacon (str, optional): The prefix for each line. Defaults to "".
            verbose (bool, optional): Whether to print the string. Defaults to True.

        Returns:
            str: The string with the history information.
        """
        string = ""
        string += f"{beacon}Space ({self.pk=}):\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.name=}\n"
        string += f"{beacon}\t{self.uuid=}\n"
        string += f"{beacon}Exchange Account:\n"
        new_beacon = beacon + "\t"
        string += f"{self.exchange_account.info(verbose=False, beacon=new_beacon)}\n"

        if verbose:  # pragma: no cover
            print(string)
        return string

    @property
    def testing(self) -> bool:
        """Testing property."""
        return self.exchange_account.testing

    @property
    def value(self) -> float:
        """Return quick value of space's wallet."""
        stablecoins = STABLECOINS.get(self.exchange_account.exchange.name)

        def _add_currency_value(currency: Currency) -> float:
            if currency.ticker in stablecoins:
                return currency.amount
            # controller = Controller.objects.filter(pair=f"{currency.ticker}USDT").first()  # noqa
            controller = Controller.get(exchange_account=self.exchange_account, base=currency.ticker, quote="USDT")
            return controller.get_price() * currency.amount
            # return controller.price * currency.amount

        value = 0

        # Connections values
        connections = self.wallet.connections.all()
        for connection in connections:
            for currency in connection.wallet.currencies.all():
                value += _add_currency_value(currency)

        # Wallet value
        for currency in self.wallet.currencies.all():
            value += _add_currency_value(currency)
        return value

    @property
    def fleets(self) -> models.QuerySet:
        """Fleets of the space."""
        connections = self.wallet.connections.all()
        return Fleet.objects.filter(clusters__links__bot__connections__in=connections).distinct()

    @property
    def bots(self) -> models.QuerySet:
        """Bots of the space."""
        connections = self.wallet.connections.all()
        return Bot.objects.filter(connections__in=connections)

    def get_stats(self) -> dict[str, int | float | str]:
        """Statistics of space."""
        order_count_30 = Order.objects.filter(
            connection__in=self.wallet.connections.all(),
            created_at__gt=datetime.now(tz=get_default_timezone()) - timedelta(days=30),
        ).count()

        return {
            "value": self.value,
            "order_count_30": order_count_30,
            "delta_30": 0,  # need history on space's value
        }

    def delete(self) -> None:
        """Delete space."""
        if self.testing:
            return super().delete()
        if self.value > 0:
            raise SpaceError.DeleteError
        return super().delete()

    def invest(self, amount: float, ticker: str) -> Credit:
        """Invest in space."""
        if ticker not in EXCHANGE_TICKERS.get("BINANCE"):
            error_msg: str = f"{ticker} is not available on {self.exchange_account.name} exchange."
            raise ExchangeError.UnavailableTicker(error_msg)

        # Real invest
        if not self.testing:
            error_msg: str = "Investing for real is not available yet."
            raise NotImplementedError(error_msg)

        # Testing invest
        Credit.objects.create(wallet=self.wallet, amount=amount, ticker=ticker)

    def withdraw(self, amount: float, ticker: str) -> Debit:
        """Withdraw from space."""
        if ticker not in [currency.ticker for currency in self.wallet.currencies.all()]:
            error_msg: str = f"{ticker} is not on your {self.name}(space)'s wallet."
            raise WalletError.UnavailableTicker(error_msg)

        # Real withdraw
        if not self.testing:
            error_msg: str = "Withdrawing for real is not available yet."
            raise NotImplementedError(error_msg)

        # Testing withdraw
        Debit.objects.create(
            wallet=self.wallet,
            amount=amount,
            ticker=ticker,
        )



from django.db.utils import IntegrityError

from django_napse.core.models import Bot, Controller, Credit, EmptyBotConfig, EmptyStrategy, Fleet, SinglePairArchitecture, Space
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.db.accounts.test_space -v2 --keepdb --parallel
"""


class SpaceTestCase:
    model = Space

    def simple_create(self):
        return Space.objects.create(name="Test Space", exchange_account=self.exchange_account, description="This is a test space")

    def test_error_create_napse_space_with_same_name(self):
        Space.objects.create(name="Test Space", exchange_account=self.exchange_account, description="This is a test space")
        with self.assertRaises(IntegrityError):
            Space.objects.create(name="Test Space", exchange_account=self.exchange_account, description="This is a test space")

    def _build_fleet_context(self):
        config = EmptyBotConfig.objects.create(space=self.space, settings={"empty": True})
        controller = Controller.get(
            exchange_account=self.exchange_account,
            base="BTC",
            quote="USDT",
            interval="1m",
        )
        architecture = SinglePairArchitecture.objects.create(constants={"controller": controller})
        strategy = EmptyStrategy.objects.create(config=config, architecture=architecture)
        template_bot = Bot.objects.create(name="Test Bot", strategy=strategy)
        fleet = Fleet.objects.create(
            name="Test Fleet",
            exchange_account=self.exchange_account,
            clusters=[
                {
                    "template_bot": template_bot,
                    "share": 0.7,
                    "breakpoint": 1000,
                    "autoscale": False,
                },
                {
                    "template_bot": template_bot,
                    "share": 0.3,
                    "breakpoint": 1000,
                    "autoscale": True,
                },
            ],
        )
        Credit.objects.create(wallet=self.space.wallet, amount=1000, ticker="USDT")
        connections = fleet.invest(self.space, 1000, "USDT")
        return fleet, connections

    def test_value_property(self):
        self._build_fleet_context()
        self.assertEqual(self.space.value, 1000)

    def test_fleets_property(self):
        fleet, _ = self._build_fleet_context()
        self.assertEqual(len(self.space.fleets), 1)
        self.assertEqual(self.space.fleets[0].uuid, fleet.uuid)


class NaspeSpaceBINANCETestCase(SpaceTestCase, ModelTestCase):
    exchange = "BINANCE"



# Generated by Django 4.2.6 on 2023-10-17 15:23

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_napse_core", "0005_exchangeaccount_default"),
    ]

    operations = [
        migrations.AlterField(
            model_name="exchangeaccount",
            name="name",
            field=models.CharField(max_length=200),
        ),
    ]



"""Django settings for napse project.

Generated by 'django-admin startproject' using Django 3.2.15.

For more information on this file, see
https://docs.djangoproject.com/en/3.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.2/ref/settings/
"""

from pathlib import Path

import environ

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent
env = environ.Env()
env.read_env(str(BASE_DIR / ".env"))
# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-secret"  # noqa: S105

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ["localhost", "127.0.0.1", "django"]
CORS_ALLOWED_ORIGINS = [f"http://{host}:8888" for host in ALLOWED_HOSTS]
CSRF_TRUSTED_ORIGINS = [f"http://{host}:8888" for host in ALLOWED_HOSTS]

# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    # Utils
    "django_extensions",
    # Napse
    "django_napse.core",
    "django_napse.simulations",
    "django_napse.auth",
    # Required
    "django_celery_beat",
    "corsheaders",
    "rest_framework",
    "rest_framework_api_key",
    "drf_spectacular",
]

MIDDLEWARE = [
    "debug_toolbar.middleware.DebugToolbarMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "corsheaders.middleware.CorsMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "tests.test_app.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "wsgi.application"


# Database
# https://docs.djangoproject.com/en/3.2/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    },
}


# Password validation
# https://docs.djangoproject.com/en/3.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.2/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.2/howto/static-files/

STATIC_URL = "/static/"

# Default primary key field type
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

# Napse settings

NAPSE_SECRETS_FILE_PATH = BASE_DIR / "secrets.json"
NAPSE_ENV_FILE_PATH = BASE_DIR / ".env"
CELERY_BROKER_URL = "redis://localhost:6379"
CELERY_BEAT_SCHEDULER = "django_celery_beat.schedulers.DatabaseScheduler"

REST_FRAMEWORK = {
    "DEFAULT_PERMISSION_CLASSES": [
        "django_napse.api.custom_permissions.HasAdminPermission",
    ],
    "DEFAULT_SCHEMA_CLASS": "drf_spectacular.openapi.AutoSchema",
}

SPECTACULAR_SETTINGS = {
    "TITLE": "Django Napse API",
    "DESCRIPTION": "List and description of all endpoints. This API documentation is essentially generated automatically. "
    "It may therefore contain errors. If you see an error, please open an issue, and if you would like "
    "to improve it, please make a contribution.",
    "VERSION": "0.0.1",
    "SERVE_INCLUDE_SCHEMA": False,
    # OTHER SETTINGS
}



from datetime import datetime, timedelta

from binance.helpers import interval_to_milliseconds
from django.db import models
from pytz import UTC

from django_napse.core.models.bots.architecture import Architecture


class SinglePairArchitecture(Architecture):
    controller = models.ForeignKey("Controller", on_delete=models.CASCADE, related_name="single_pair_architectures")
    variable_last_candle_date = models.DateTimeField(default=datetime(1970, 1, 1, 0, 0, 0, 0, tzinfo=UTC))

    def __str__(self) -> str:
        return f"SINGLE PAIR ARCHITECHTURE {self.pk}"

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}Single Pair Architecture {self.pk}:\n"
        new_beacon = beacon + "\t"
        string += f"{beacon}Controller:\n{beacon}{self.controller.info(beacon=new_beacon, verbose=False)}\n"

        if verbose:  # pragma: no cover
            print(string)
        return string

    def copy(self):
        return SinglePairArchitecture.objects.create(constants={"controller": self.controller}, variables={})

    def controllers_dict(self):
        return {"main": self.controller}

    def skip(self, data: dict) -> bool:
        if data["candles"][data["controllers"]["main"]]["current"]["open_time"] < self.variable_last_candle_date + timedelta(
            milliseconds=interval_to_milliseconds(data["controllers"]["main"].interval),
        ):
            return True
        return False

    def architecture_modifications(self, order: dict, data: dict):
        return [
            {
                "key": "last_candle_date",
                "value": str(data["candles"][data["controllers"]["main"]]["current"]["open_time"]),
                "target_type": "datetime",
                "ignore_failed_order": True,
            },
        ]

    def accepted_tickers(self):
        return [self.controller.base, self.controller.quote]

    def accepted_investment_tickers(self):
        return [self.controller.quote]



# Generated by Django 4.2.5 on 2023-10-09 15:35

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("django_napse_core", "0001_initial"),
        ("rest_framework_api_key", "0005_auto_20220110_1102"),
    ]

    operations = [
        migrations.CreateModel(
            name="KeyPermission",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                ("approved", models.BooleanField(default=False)),
                ("revoked", models.BooleanField(default=False)),
                ("permission_type", models.CharField(max_length=200)),
                (
                    "key",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="permissions",
                        to="rest_framework_api_key.apikey",
                    ),
                ),
                (
                    "space",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_napse_core.napsespace",
                    ),
                ),
            ],
            options={
                "unique_together": {("key", "space", "permission_type")},
            },
        ),
    ]



from django_napse.core.models import Credit, Debit
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.transactions.test_debit -v2 --keepdb --parallel
"""


class DebitTestCase:
    model = Debit

    def simple_create(self):
        Credit.objects.create(wallet=self.space.wallet, amount=10, ticker="BTC")
        return Debit.objects.create(
            wallet=self.space.wallet,
            amount=1,
            ticker="BTC",
        )

    def test_empty_credit(self):
        debit = Debit.objects.create(
            wallet=self.space.wallet,
            amount=0,
            ticker="BTC",
        )
        self.assertIsNone(debit)

    def test_correct_amount(self):
        Credit.objects.create(wallet=self.space.wallet, amount=10, ticker="BTC")
        Debit.objects.create(
            wallet=self.space.wallet,
            amount=1,
            ticker="BTC",
        )
        self.assertEqual(self.space.wallet.get_amount("BTC"), 9)
        Debit.objects.create(
            wallet=self.space.wallet,
            amount=1,
            ticker="BTC",
        )
        self.assertEqual(self.space.wallet.get_amount("BTC"), 8)


class DebitBINANCETestCase(DebitTestCase, ModelTestCase):
    exchange = "BINANCE"



from django_napse.core.models import Bot, Controller, Credit, EmptyBotConfig, EmptyStrategy, Fleet, SinglePairArchitecture, Space
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.db.fleets.test_fleet -v2 --keepdb --parallel
"""


class FleetTestCase:
    model = Fleet

    def simple_create(self):
        config = EmptyBotConfig.objects.create(space=self.space, settings={"empty": True})
        controller = Controller.get(
            exchange_account=self.exchange_account,
            base="BTC",
            quote="USDT",
            interval="1m",
        )
        architecture = SinglePairArchitecture.objects.create(constants={"controller": controller})
        strategy = EmptyStrategy.objects.create(config=config, architecture=architecture)
        bot = Bot.objects.create(name="Test Bot", strategy=strategy)
        return self.model.objects.create(
            name="Test Fleet",
            exchange_account=self.exchange_account,
            clusters=[
                {
                    "template_bot": bot,
                    "share": 0.7,
                    "breakpoint": 1000,
                    "autoscale": False,
                },
                {
                    "template_bot": bot,
                    "share": 0.3,
                    "breakpoint": 1000,
                    "autoscale": True,
                },
            ],
        )

    def test_invest(self):
        Credit.objects.create(wallet=self.space.wallet, amount=1000, ticker="USDT")
        fleet = self.simple_create()
        connections = fleet.invest(self.space, 1000, "USDT")
        self.assertEqual(self.space.wallet.get_amount("USDT"), 0)
        self.assertEqual(connections[0].wallet.get_amount("USDT"), 700)
        self.assertEqual(connections[1].wallet.get_amount("USDT"), 300)

    def test_invest_twice(self):
        Credit.objects.create(wallet=self.space.wallet, amount=2000, ticker="USDT")
        fleet = self.simple_create()
        fleet.invest(self.space, 1000, "USDT")
        connections = fleet.invest(self.space, 1000, "USDT")
        self.assertEqual(self.space.wallet.get_amount("USDT"), 0)
        self.assertEqual(connections[0].wallet.get_amount("USDT"), 1400)
        self.assertEqual(connections[1].wallet.get_amount("USDT"), 600)

    def _two_spaces_on_same_fleet(self):
        fleet = self.simple_create()

        space = Space.objects.create(name="Random Space", exchange_account=self.exchange_account, description="This is a test space")
        Credit.objects.create(wallet=space.wallet, amount=2000, ticker="USDT")
        fleet.invest(space, 1000, "USDT")

        Credit.objects.create(wallet=self.space.wallet, amount=2000, ticker="USDT")
        fleet.invest(self.space, 1000, "USDT")

        return fleet

    def test_value(self):
        Credit.objects.create(wallet=self.space.wallet, amount=1000, ticker="USDT")
        fleet = self.simple_create()
        fleet.invest(self.space, 1000, "USDT")
        self.assertEqual(fleet.value, 1000)

    def test_value_with_two_spaces(self):
        fleet = self._two_spaces_on_same_fleet()
        self.assertEqual(fleet.value, 2000)

    def test_space_frame_value(self):
        Credit.objects.create(wallet=self.space.wallet, amount=1000, ticker="USDT")
        fleet = self.simple_create()
        fleet.invest(self.space, 1000, "USDT")
        self.assertEqual(fleet.space_frame_value(self.space), 1000)

    def test_space_frame_value_with_two_spaces(self):
        fleet = self._two_spaces_on_same_fleet()
        self.assertEqual(fleet.space_frame_value(self.space), 1000)


class FleetBINANCETestCase(FleetTestCase, ModelTestCase):
    exchange = "BINANCE"



from typing import TYPE_CHECKING

from django.db import models

from django_napse.core.models.connections.connection import Connection
from django_napse.core.models.wallets.wallet import Wallet

if TYPE_CHECKING:
    from django_napse.core.models.accounts.exchange import ExchangeAccount
    from django_napse.core.models.accounts.space import Space
    from django_napse.core.models.bots.bot import Bot


class SpaceWallet(Wallet):
    """A Wallet owned by a Space."""

    owner = models.OneToOneField("Space", on_delete=models.CASCADE, related_name="wallet")

    def __str__(self) -> str:  # pragma: no cover
        return f"WALLET: {self.pk=}. OWNER: {self.owner=}"

    @property
    def space(self) -> "Space":
        """Return the space that owns the wallet."""
        return self.owner

    @property
    def exchange_account(self) -> "ExchangeAccount":
        """Return the exchange account that contains the wallet."""
        return self.space.exchange_account.find()

    def connect_to_bot(self, bot: "Bot") -> "Connection":
        """Get or create connection to bot.

        Args:
            bot (Bot): The bot to connect to.

        Returns:
            Connection: The connection to the bot.
        """
        try:
            connection = self.connections.get(owner=self, bot=bot)
        except Connection.DoesNotExist:
            connection = Connection.objects.create(owner=self, bot=bot)
        return connection



class SimulationError:
    """Base class for simulation errors."""

    class InvalidData(Exception):
        """Raised when invalid data is passed."""

    class BotSimQueueError(Exception):
        """Raised when an error occurs in a BotSimQueue."""

    class CancelledSimulationError(Exception):
        """Raised when a simulation is cancelled."""


class DataSetError:
    """Base class for DataSet errors."""

    class DataSetQueueError(Exception):
        """Raised when an error occurs in a DataSetQueue."""

    class InvalidSettings(Exception):
        """Raised when an invalid setting is passed."""



import contextlib

from django.db import IntegrityError, models
from rest_framework_api_key.models import APIKey

from django_napse.auth.models.permissions import KeyPermission
from django_napse.utils.errors import NapseKeyError


class NapseAPIKey(APIKey):
    """Napse API Key is use for authentication."""

    is_master_key = models.BooleanField(default=False)
    description = models.TextField(blank=True)

    def save(self, *args: list, **kwargs: dict[str, any]) -> None:
        """Save the instance."""
        if self.is_master_key:
            self.name = "Napse Master Key"
        if self.is_master_key and NapseAPIKey.objects.filter(is_master_key=True).count() > (1 if self.pk else 0):
            error_msg = "Only one master key can exist at a time."
            raise NapseKeyError.DuplicateMasterkey(error_msg)
        return super().save(*args, **kwargs)

    def add_permission(self, space, permission):
        """Add a permission on space to the key."""
        with contextlib.suppress(IntegrityError):
            KeyPermission.objects.create(key=self, space=space, permission_type=permission)

    def revoke(self) -> None:
        """Revoque the key."""
        self.revoked = True
        self.save()



from django.apps import apps
from django.db import models


class ExchangeAccountManager(models.Manager):
    def create(
        self,
        exchange,
        testing: bool,
        name: str,
        description: str = "",
        **kwargs,
    ):
        Space = apps.get_model("django_napse_core", "Space")
        exchange_account = self.model(
            exchange=exchange,
            name=name,
            description=description,
            testing=testing,
            **kwargs,
        )
        exchange_account.save()
        Space.objects.create(
            name=exchange_account.name,
            description=f"This is the main space for the {exchange_account.name} exchange account.",
            exchange_account=exchange_account,
        )
        return exchange_account



from typing import ClassVar

from rest_framework import serializers

from django_napse.api.bots.serializers import BotSerializer
from django_napse.core.models import Bot, Cluster


class ClusterSerializerV1(serializers.ModelSerializer):
    """Serializer for Cluster model."""

    template_bot = BotSerializer()

    class Meta:  # noqa: D106
        model = Cluster
        fields: ClassVar[list[str]] = [
            "template_bot",
            "share",
            "breakpoint",
            "autoscale",
        ]


class ClusterFormatterSerializer(serializers.Serializer):
    """Format cluster dictionnary for fleet creation."""

    template_bot = serializers.UUIDField(required=True)
    share = serializers.FloatField(required=True)
    breakpoint = serializers.IntegerField(required=True)
    autoscale = serializers.BooleanField(required=True)

    def validate(self, data: dict[str, any]) -> dict[str, any]:
        """Validate the data & get template bot instance."""
        data = super().validate(data)
        try:
            bot = Bot.objects.get(uuid=data.pop("template_bot"))
        except Bot.DoesNotExist:
            error_msg: str = "Template bot does not exist."
            raise serializers.ValidationError(error_msg) from None
        data["template_bot"] = bot
        return data



from __future__ import annotations

from rest_framework import serializers

from django_napse.core.models.histories.history import HistoryDataPointField


class HistoryDataPointFieldSerializer(serializers.ModelSerializer):
    """Serialize a wallet instance."""

    class Meta:  # noqa: D106
        model = HistoryDataPointField
        fields: str = ["key", "value", "target_type"]



from django.apps import apps
from django.db import models


class ConnectionManager(models.Manager):
    """Define how the Connection.objects works."""

    def create(self, owner: "Wallet", bot: "Bot") -> "Connection":  # noqa: F821
        ConnectionWallet = apps.get_model("django_napse_core", "ConnectionWallet")

        connection = self.model(owner=owner, bot=bot)
        connection.save()
        # bot.init_specific_args(connection)  # noqa: ERA001
        ConnectionWallet.objects.create(
            title=f"Connection between {owner.title} - {bot.name}",
            owner=connection,
        )
        return connection



from django_napse.core.models import Bot, Connection, Controller, EmptyBotConfig, EmptyStrategy, Order, OrderBatch, SinglePairArchitecture
from django_napse.utils.constants import SIDES
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.orders.test_order -v2 --keepdb --parallel
"""


class OrderTestCase:
    model = Order

    def simple_create(self):
        config = EmptyBotConfig.objects.create(space=self.space, settings={"empty": True})
        controller = Controller.get(
            exchange_account=self.exchange_account,
            base="BTC",
            quote="USDT",
            interval="1m",
        )
        architecture = SinglePairArchitecture.objects.create(constants={"controller": controller})
        strategy = EmptyStrategy.objects.create(config=config, architecture=architecture)
        bot = Bot.objects.create(name="Test Bot", strategy=strategy)
        connection = Connection.objects.create(owner=self.space.wallet, bot=bot)
        connection.wallet.top_up(amount=10000, ticker="USDT", force=True)
        batch = OrderBatch.objects.create(controller=controller)
        return Order.objects.create(
            batch=batch,
            connection=connection,
            asked_for_amount=10,
            asked_for_ticker="USDT",
            price=30000,
            pair="BTCUSDT",
            side=SIDES.BUY,
        )


class OrderBINANCETestCase(OrderTestCase, ModelTestCase):
    exchange = "BINANCE"



from typing import TYPE_CHECKING

from django.db import models

from django_napse.core.models.histories.history import History, HistoryDataPoint
from django_napse.utils.constants import HISTORY_DATAPOINT_FIELDS, HISTORY_DATAPOINT_FIELDS_WILDCARDS

if TYPE_CHECKING:
    from django_napse.core.models.wallets.wallet import Wallet


class WalletHistory(History):
    """A History for a Wallet.

    Use it to track the evolution of a bot over time.

    This tracks the following fields:
    - `value`: The value of the wallet in USD.
    - `amount_ticker`: For each ticker, the amount of the ticker in the wallet.
    """

    owner: "Wallet" = models.OneToOneField("Wallet", on_delete=models.CASCADE, related_name="history")

    def generate_data_point(self) -> "HistoryDataPoint":
        """Create a new data point for the bot."""
        wallet = self.owner.to_dict()
        points = {
            HISTORY_DATAPOINT_FIELDS.WALLET_VALUE: self.owner.value_market(),
        }
        for currency in wallet.currencies.values():
            points[HISTORY_DATAPOINT_FIELDS_WILDCARDS.AMOUNT + currency.ticker] = currency.amount

        return HistoryDataPoint.objects.create(
            history=self,
            points=points,
        )



from django.db import models

from django_napse.utils.constants import MODIFICATION_STATUS
from django_napse.utils.findable_class import FindableClass
from django_napse.utils.usefull_functions import process_value_from_type


class Modification(models.Model, FindableClass):
    """A modification to be applied to an order."""

    order = models.ForeignKey("Order", on_delete=models.CASCADE, related_name="modifications")

    status = models.CharField(default=MODIFICATION_STATUS.PENDING, max_length=15)
    created_at = models.DateTimeField(auto_now_add=True)
    ignore_failed_order = models.BooleanField(default=False)

    key = models.CharField(max_length=100)
    value = models.CharField(max_length=100)
    target_type = models.CharField(max_length=100)

    def __str__(self) -> str:
        return f"MODIFICATION: {self.pk=}"

    def get_value(self, **kwargs: dict[str, any]) -> any:
        """Return value into classic python object."""
        return process_value_from_type(self.value, self.target_type, **kwargs)

    def apply__no_db(self) -> tuple[models.Model, "Modification"]:
        """Apply the modification without saving it to the database.

        Raises:
            NotImplementedError: You must implement this method in your subclass.
        """
        error_msg = "You must implement this method in your subclass"
        raise NotImplementedError(error_msg)

    def apply(self) -> None:
        """Apply the modification to the order and save them to the database.

        Raises:
            NotImplementedError: You must implement this method in your subclass.
        """
        error_msg = "You must implement this method in your subclass"
        raise NotImplementedError(error_msg)



from datetime import datetime

from django.apps import apps
from django.db import models


class SimulationQueueManager(models.Manager):
    def create(
        self,
        space,
        bot,
        start_date: datetime,
        end_date: datetime,
        investments: dict,
    ):
        SimulationQueueInvestedCurrency = apps.get_model("django_napse_simulations", "SimulationQueueInvestedCurrency")
        queue = self.model(
            space=space,
            bot=bot,
            start_date=start_date,
            end_date=end_date,
        )
        queue.save()
        for ticker, investment in investments.items():
            SimulationQueueInvestedCurrency.objects.create(
                owner=queue,
                ticker=ticker,
                amount=investment,
            )
        return queue



from typing import ClassVar
from uuid import UUID

from rest_framework import serializers

from django_napse.api.orders.serializers import OrderSerializer
from django_napse.api.wallets.serializers import WalletSerializer
from django_napse.core.models import Bot, BotHistory, ConnectionWallet, Order, Space


class BotSerializer(serializers.ModelSerializer):
    """Serialize bot instances."""

    delta = serializers.SerializerMethodField(read_only=True)
    space = serializers.SerializerMethodField(read_only=True)
    value = serializers.SerializerMethodField(read_only=True)
    exchange_account = serializers.SerializerMethodField(read_only=True)
    fleet = serializers.CharField(source="fleet.uuid", read_only=True)

    class Meta:  # noqa: D106
        model = Bot
        fields: ClassVar = [
            "name",
            "uuid",
            "value",
            "delta",
            "fleet",
            "space",
            "exchange_account",
        ]
        read_only_fields: ClassVar = [
            "uuid",
            "value",
            "delta",
            "space",
        ]

    def __init__(self, *args: list[any], space: Space | None = None, **kwargs: dict[str, any]) -> None:
        """Add space to the serializer and run the default constructor."""
        self.space = space
        super().__init__(*args, **kwargs)

    def get_delta(self, instance: Bot) -> float:
        """Delta on the last 30 days."""
        try:
            history = BotHistory.objects.get(owner=instance)
        except BotHistory.DoesNotExist:
            return 0
        return history.get_delta()

    def get_space(self, instance: Bot) -> UUID | None:  # noqa: ARG002
        """Return the space used for the space containerization."""
        if self.space is None:
            return None
        print("UUID", type(self.space.uuid))
        return self.space.uuid

    def get_exchange_account(self, instance: Bot) -> UUID | None:
        """Return the exchange account of the bot if it exists."""
        if not instance.is_in_fleet and not instance.is_in_simulation:
            return None
        return instance.exchange_account.uuid

    def get_value(self, instance: Bot) -> float:
        """Return bot's value."""
        return instance.value(space=self.space)

    def create(self, validated_data: dict[str, any]) -> Bot:  # noqa: ARG002
        """Create a bot instance."""
        error_msg: str = "You don't have to create a bot from an endpoint"
        raise ValueError(error_msg)


class BotDetailSerializer(serializers.ModelSerializer):
    """Deep dive in bot's data for serialization."""

    delta = serializers.SerializerMethodField(read_only=True)
    space = serializers.SerializerMethodField(read_only=True)
    value = serializers.SerializerMethodField(read_only=True)
    exchange_account = serializers.CharField(source="exchange_account.uuid", read_only=True)
    fleet = serializers.CharField(source="fleet.uuid", read_only=True)

    statistics = serializers.SerializerMethodField(read_only=True)
    wallet = serializers.SerializerMethodField(read_only=True)
    orders = serializers.SerializerMethodField(read_only=True)

    class Meta:  # noqa: D106
        model = Bot
        fields: ClassVar[list[str]] = [
            "name",
            "uuid",
            "value",
            "delta",
            "statistics",
            "fleet",
            "space",
            "exchange_account",
            "wallet",
            "orders",
        ]
        read_only_fields: ClassVar = fields

    def __init__(self, *args: list[any], space: Space | None = None, **kwargs: dict[str, any]) -> None:
        """Add space to the serializer and run the default constructor."""
        self.space = space
        super().__init__(*args, **kwargs)

    def get_delta(self, instance: Bot) -> float:
        """Delta on the last 30 days."""
        try:
            history = BotHistory.objects.get(owner=instance)
        except BotHistory.DoesNotExist:
            return 0
        return history.get_delta()

    def get_value(self, instance: Bot) -> float:
        """Return bot's value."""
        return instance.value(space=self.space)

    def get_space(self, instance: Bot) -> UUID | None:  # noqa: ARG002
        """Return the space used for the space containerization."""
        if self.space is None:
            return None
        return self.space.uuid

    def get_statistics(self, instance: Bot) -> dict[str, str | int | float]:
        """Return Bot's statistics."""
        return instance.get_stats(space=self.space)

    def get_wallet(self, instance: Bot) -> dict[str, any] | list[dict[str, any]]:
        """Return space's connections wallet of this bot, or return all connections wallets."""
        if self.space is not None:
            wallet = ConnectionWallet.objects.get(
                owner__owner=self.space.wallet,
                owner__bot=instance,
            )
            return WalletSerializer(wallet).data

        wallets = [connection.wallet for connection in instance.connections.all()]
        return WalletSerializer(wallets, many=True).data

    def get_orders(self, instance: Bot) -> list[dict[str, any]]:
        """Return all orders of the bot."""
        if self.space is None:
            return OrderSerializer(
                Order.objects.filter(connection__bot=instance),
                many=True,
            ).data

        return OrderSerializer(
            Order.objects.filter(
                connection__bot=instance,
                connection__owner=self.space.wallet,
            ),
            many=True,
        ).data

    def create(self, validated_data: dict[str, any]) -> Bot:  # noqa: ARG002
        """Create a bot instance."""
        error_msg: str = "You don't have to create a bot from an endpoint"
        raise ValueError(error_msg)

    def update(self, instance: Bot, validated_data: dict[str, any]) -> Bot:  # noqa: ARG002
        """Update a bot instance."""
        error_msg: str = "You don't have to update a bot from an endpoint"
        raise ValueError(error_msg)



from django.test import TestCase

from django_napse.core.models import EXCHANGE_ACCOUNT_DICT, Space
from django_napse.core.settings import napse_settings


class CustomTestCase(TestCase):
    exchange = None

    def skip_condition(self):
        if self.__class__.__name__ == "ModelTestCase":
            return True, "because ModelTestCase (base class) tests shouldn't be run."
        if self.__class__.__name__ == "APITestCase":
            return True, "because ModelTestCase (base class) tests shouldn't be run."

        if not self.__class__.__name__.endswith(f"{self.exchange}TestCase"):
            error_msg = f"CustomTestCase subclass {self.__class__.__name__} name must end with <Exchange>TestCase."
            error_msg += " Make sure you are writing tests as specified in the docs."
        if self.exchange is not None and self.exchange not in napse_settings.NAPSE_EXCHANGES_TO_TEST:
            return True, f"because {self.exchange} is not in {napse_settings.NAPSE_EXCHANGES_TO_TEST}"
        return False, ""

    @property
    def exchange_account(self):
        try:
            return EXCHANGE_ACCOUNT_DICT[self.exchange].objects.get(name=f"{self.exchange} Test Account")
        except KeyError as e:
            error_msg = f"Exchange {self.exchange} isn't a valid choice."
            raise ValueError(error_msg) from e

    @property
    def space(self):
        try:
            return Space.objects.get(name="Test Space", exchange_account=self.exchange_account, description="This is a test space")
        except Space.DoesNotExist:
            return Space.objects.create(name="Test Space", exchange_account=self.exchange_account, description="This is a test space")



# Generated by Django 4.2.5 on 2023-10-05 16:16

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("django_napse_core", "0002_history_historydatapoint_historydatapointfield"),
    ]

    operations = [
        migrations.CreateModel(
            name="WalletHistory",
            fields=[
                (
                    "history_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.history",
                    ),
                ),
                (
                    "owner",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="history",
                        to="django_napse_core.wallet",
                    ),
                ),
            ],
            bases=("django_napse_core.history",),
        ),
    ]



from typing import TYPE_CHECKING

from django.db import models

from django_napse.core.models.wallets.wallet import Wallet

if TYPE_CHECKING:
    from django_napse.core.models.accounts.exchange import ExchangeAccount


class OrderWallet(Wallet):
    """A Wallet owned by an Order."""

    owner = models.OneToOneField("Order", on_delete=models.CASCADE, related_name="wallet")

    def __str__(self) -> str:  # pragma: no cover
        return f"WALLET: {self.pk=}. OWNER: {self.owner=}"

    @property
    def exchange_account(self) -> "ExchangeAccount":
        """Return the exchange account that contains the wallet."""
        return self.owner.exchange_account.find()



# Generated by Django 4.2.7 on 2024-01-07 13:23

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_napse_core", "0007_historydatapoint_created_at"),
    ]

    operations = [
        migrations.CreateModel(
            name="BotHistory",
            fields=[
                (
                    "history_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_napse_core.history",
                    ),
                ),
                ("owner", models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name="history", to="django_napse_core.bot")),
            ],
            bases=("django_napse_core.history",),
        ),
    ]



import re
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING, List

from rest_framework.routers import DefaultRouter
from rest_framework.viewsets import GenericViewSet

if TYPE_CHECKING:
    from types import ModuleType


class ConflictingUrlNamesError(Exception):
    """When several viewset have the same automatic url."""


def build_main_router() -> DefaultRouter:
    """Create a main router object and register the appropriate viewsets to it based on the modules and classes found in the `api` directory.

    Returns:
        DefaultRouter: The main router object with registered URL patterns.
    """
    main_router = DefaultRouter()
    url_name_list: List[str] = []

    api_dir = Path(__file__).parent
    api_modules_folders_names = [folder.name for folder in api_dir.iterdir() if folder.is_dir() and not folder.name.startswith("_")]
    for module_name in api_modules_folders_names:
        try:
            module: ModuleType = import_module(f"django_napse.api.{module_name}.views")
        except (ImportError, ModuleNotFoundError) as error:
            print(f"Could not import module {module_name} ({type(error)})")
            print(error)
            continue
        for obj in vars(module).values():
            if isinstance(obj, type) and issubclass(obj, GenericViewSet):
                # from CamelCase to snake_case & remove "_view" (ex: MyWalletView -> my_wallet)
                url_name = re.sub(r"(?<!^)(?=[A-Z])", "_", obj.__name__).lower().replace("_view", "")

                if url_name in url_name_list:
                    error_msg: str = f"Url name {url_name} already exists"
                    raise ConflictingUrlNamesError(error_msg)
                main_router.register(url_name, obj, basename=url_name)
                url_name_list.append(url_name)

    return main_router


main_api_router = build_main_router()



from django.db import models

from django_napse.core.models.bots.config import BotConfig


class EmptyBotConfig(BotConfig):
    setting_empty = models.BooleanField()

    def __str__(self) -> str:
        return f"DCA BOT CONFIG: {self.pk} - {self.immutable}"



import contextlib
import json

from django.apps import apps
from django.db.models.signals import post_migrate
from django.dispatch import receiver

from django_napse.core.settings import napse_settings
from django_napse.utils.errors import NapseKeyError


@receiver(post_migrate)
def create_master_key(sender, **kwargs):
    NapseAPIKey = apps.get_model("django_napse_auth", "NapseAPIKey")
    with contextlib.suppress(NapseKeyError.DuplicateMasterkey):
        _, key = NapseAPIKey.objects.create_key(name="", is_master_key=True)
        print(f"Master key created: {key}")
        with open(napse_settings.NAPSE_SECRETS_FILE_PATH, "r") as f:
            secrets = json.load(f)
            secrets["master_key"] = key
        with open(napse_settings.NAPSE_SECRETS_FILE_PATH, "w") as f:
            json.dump(secrets, f, indent=4)



# from django_napse.core.models import Order
from django_napse.core.tasks.base_task import BaseTask


class OrderProcessExecutorTask(BaseTask):
    name = "order_process_executor"
    interval_time = 1  # Impossible to make dynamic modification because of celery

    def run(self) -> None:
        """Run a task to process all pending orders."""
        if not self.avoid_overlap(verbose=False):
            return
        processed = 0
        # for order in Order.objects.filter(status="pending", completed=True):
        #     processed += 1
        if processed > 0:
            self.info(f"Processed {processed} orders")


OrderProcessExecutorTask().delete_task()
OrderProcessExecutorTask().register_task()



from django.db.utils import IntegrityError
from django.test import TestCase

from django_napse.core.models import Currency, CurrencyPydantic, SpaceWallet, Wallet, WalletPydantic
from django_napse.utils.errors import WalletError
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.wallets.test_wallet -v2 --keepdb --parallel
"""


class WalletTestCase(TestCase):
    def test_cant_create_wallet(self):
        with self.assertRaises(WalletError.CreateError):
            Wallet.objects.create(title="Test Wallet", owner=None)


class BaseWalletTestCase:
    def simple_create(self):
        return self.owner.wallet

    def test_error_duplicate_wallet(self):
        with self.assertRaises(IntegrityError):
            self.model.objects.create(title="Test Wallet", owner=self.owner)

    def test_property_testing(self):
        wallet = self.simple_create()
        self.assertTrue(wallet.testing)

    # Spend
    def test_spend_security(self):
        wallet = self.simple_create()
        with self.assertRaises(WalletError.SpendError):
            wallet.spend(ticker="BTC", amount=0.5)

    def test_force_spend(self):
        wallet = self.simple_create()
        Currency.objects.create(wallet=wallet, ticker="BTC", amount=1, mbp=20000)
        wallet.spend(ticker="BTC", amount=0.5, force=True)

    def test_insufficient_funds(self):
        wallet = self.simple_create()
        Currency.objects.create(wallet=wallet, ticker="BTC", amount=1, mbp=20000)
        with self.assertRaises(WalletError.SpendError):
            wallet.spend(ticker="BTC", amount=2, force=True)

    def test_non_existing_funds(self):
        wallet = self.simple_create()
        Currency.objects.create(wallet=wallet, ticker="BTC", amount=1, mbp=20000)
        with self.assertRaises(WalletError.SpendError):
            wallet.spend(ticker="ETH", amount=2, force=True)

    def test_spend_negative(self):
        wallet = self.simple_create()
        with self.assertRaises(ValueError):
            wallet.spend(ticker="BTC", amount=-2, force=True)

    def test_error_spend_locked(self):
        wallet = self.simple_create()
        wallet.locked = True
        wallet.save()
        with self.assertRaises(TimeoutError):
            wallet.spend(ticker="BTC", amount=2, force=True)

    # Top Up
    def test_topup_security(self):
        wallet = self.simple_create()
        Currency.objects.create(wallet=wallet, ticker="BTC", amount=1, mbp=20000)
        with self.assertRaises(WalletError.TopUpError):
            wallet.top_up(ticker="BTC", amount=0.5)

    def test_force_topup(self):
        wallet = self.simple_create()
        Currency.objects.create(wallet=wallet, ticker="BTC", amount=1, mbp=20000)
        wallet.top_up(ticker="BTC", amount=0.5, mbp=1, force=True)

    def test_error_topup_locked(self):
        wallet = self.simple_create()
        wallet.locked = True
        wallet.save()
        with self.assertRaises(TimeoutError):
            wallet.top_up(ticker="BTC", amount=2, mbp=1, force=True)

    def test_topup_negative(self):
        wallet = self.simple_create()
        with self.assertRaises(ValueError):
            wallet.top_up(ticker="BTC", amount=-2, mbp=1, force=True)

    # Other
    def test_exchange_account(self):
        wallet = self.simple_create()
        self.assertEqual(wallet.exchange_account, self.owner.exchange_account)

    def test_has_funds(self):
        wallet = self.simple_create()
        self.assertFalse(wallet.has_funds(ticker="BTC", amount=1))
        Currency.objects.create(wallet=wallet, ticker="BTC", amount=1, mbp=20000)
        self.assertTrue(wallet.has_funds(ticker="BTC", amount=1))

    def test_get_amout(self):
        wallet = self.simple_create()
        self.assertEqual(wallet.get_amount(ticker="BTC"), 0)
        Currency.objects.create(wallet=wallet, ticker="BTC", amount=1, mbp=20000)
        self.assertEqual(wallet.get_amount(ticker="BTC"), 1)

    def test_value_mbp(self):
        wallet = self.simple_create()
        self.assertEqual(wallet.value_mbp(), 0)
        Currency.objects.create(wallet=wallet, ticker="BTC", amount=1, mbp=20000)
        Currency.objects.create(wallet=wallet, ticker="ETH", amount=0, mbp=1000)
        self.assertEqual(wallet.value_mbp(), 20000)

    def test_to_dict(self):
        wallet = self.simple_create()
        Currency.objects.create(wallet=wallet, ticker="BTC", amount=1, mbp=20000)
        Currency.objects.create(wallet=wallet, ticker="USDT", amount=1, mbp=1)
        self.assertEqual(
            wallet.to_dict(),
            WalletPydantic(
                title="Wallet for space Test Space",
                testing=True,
                locked=False,
                created_at=wallet.created_at,
                currencies={
                    "BTC": CurrencyPydantic(ticker="BTC", amount=1.0, mbp=20000.0),
                    "USDT": CurrencyPydantic(ticker="USDT", amount=1.0, mbp=1.0),
                },
            ),
        )


class SpaceWalletTestCase(BaseWalletTestCase):
    model = SpaceWallet

    def setUp(self):
        super().setUp()
        self.owner = self.space

    def test_with_real_exchange(self):
        self.space.wallet.top_up(ticker="BTC", amount=0.5, force=True)

    def test_space(self):
        wallet = self.simple_create()
        self.assertEqual(wallet.space, self.owner)

    def test_value_market_BTC(self):  # noqa: N802
        Currency.objects.create(wallet=self.space.wallet, ticker="BTC", amount=1, mbp=20000)
        self.assertTrue(self.space.wallet.value_market() > 0)

    def test_value_market_USDT(self):  # noqa: N802
        Currency.objects.create(wallet=self.space.wallet, ticker="USDT", amount=1, mbp=1)
        self.assertEqual(self.space.wallet.value_market(), 1)

    def test_value_zero(self):
        Currency.objects.create(wallet=self.space.wallet, ticker="BTC", amount=0, mbp=1)
        Currency.objects.create(wallet=self.space.wallet, ticker="USDT", amount=0, mbp=1)
        self.assertEqual(self.space.wallet.value_market(), 0)


class SpaceWalletBINANCETestCase(SpaceWalletTestCase, ModelTestCase):
    exchange = "BINANCE"


# class OrderWalletTestCase(BaseWalletTestCase):
#     model = OrderWallet

#     def setUp(self) -> None:
#         super().setUp()
#         config = EmptyBotConfig.objects.create(space=self.space, settings={"empty": True})
#         fleet = Fleet.objects.create(name="test_fleet", configs=[config], exchange_account=self.exchange_account)
#         bot = fleet.bots.first()
#         self.owner = Order.objects.create(bot=bot, buy_amount=100, sell_amount=100, price=1)
# def test_space(self):
#     wallet = self.simple_create()
#     self.assertEqual(wallet.space, self.owner.space)


# class OrderWalletBINANCETestCase(OrderWalletTestCase, ModelTestCase):
#     exchange = "BINANCE"


# class ConnectionWalletTestCase(BaseWalletTestCase, TestCase):
#     wallet_class = ConnectionWallet

#     def setUp(self) -> None:
#         self.owner = Conne.objects.create(name="Test Space", testing=True)
#         self.owner = Conne.objects.create(name="Test Space", testing=True)
#         self.owner = Conne.objects.create(name="Test Space", testing=True)



from django_napse.core.models.bots.plugin import Plugin
from django_napse.core.models.connections.connection import ConnectionSpecificArgs
from django_napse.utils.constants import PLUGIN_CATEGORIES, SIDES


class MBPPlugin(Plugin):
    @classmethod
    def plugin_category(cls):
        return PLUGIN_CATEGORIES.POST_ORDER

    def apply__no_db(self, data: dict) -> dict:
        order = data["order"]
        if order["side"] == SIDES.BUY:
            new_mbp = f"{order['controller'].base}|{order['price']}"
            order["ConnectionModifications"] += [
                {
                    "key": "mbp",
                    "value": new_mbp,
                    "target_type": "plugin_mbp",
                    "ignore_failed_order": False,
                    "connection_specific_arg": data["connection_data"][data["connection"]]["connection_specific_args"]["mbp"],
                },
            ]
        return data

    def _connect(self, connection):
        ConnectionSpecificArgs.objects.create(connection=connection, key="mbp", value="None", target_type="float")



import uuid

from django.db import models

from django_napse.core.models.accounts.managers.exchange import ExchangeAccountManager
from django_napse.utils.constants import EXCHANGE_TICKERS
from django_napse.utils.errors import ExchangeAccountError
from django_napse.utils.findable_class import FindableClass
from django_napse.utils.trading.binance_controller import BinanceController, ExchangeController


class Exchange(models.Model):
    """Represent an exchange entity, like Binance, Kraken, ...."""

    name = models.CharField(max_length=200, unique=True)
    description = models.TextField()

    def __str__(self) -> str:
        return f"EXCHANGE: {self.name}"

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return & display info about the exchange."""
        string = ""
        string += f"{beacon}Exchange ({self.pk=}):\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.name=}\n"
        if verbose:  # pragma: no cover
            print(string)
        return string

    def get_tickers(self) -> list[str]:
        """Return a list of available tickers on the exchange."""
        return EXCHANGE_TICKERS.get(self.name, [])


class ExchangeAccount(models.Model, FindableClass):
    """Represent an account on an exchange."""

    uuid = models.UUIDField(unique=True, editable=False, default=uuid.uuid4)

    exchange = models.ForeignKey(
        "Exchange",
        on_delete=models.CASCADE,
        related_name="accounts",
    )
    name = models.CharField(max_length=200)
    testing = models.BooleanField(default=True)
    description = models.TextField()
    default = models.BooleanField(default=False)
    created_at = models.DateTimeField(auto_now_add=True)

    objects = ExchangeAccountManager()

    def __str__(self) -> str:
        return f"EXCHANGE ACCOUNT ({self.__class__.__name__}): {self.pk=} - {self.name=} - {self.testing=}"

    def info(self, beacon: str = "", *, verbose: bool = True) -> str:
        """Return & display info about the exchange account."""
        string = ""
        string += f"{beacon}{self.__class__.__name__} ({self.pk=}):\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.name=}\n"
        string += f"{beacon}\t{self.testing=}\n"
        string += f"{beacon}\t{self.description=}\n"
        exchange_str = self.exchange.info(verbose=False, beacon=f"{beacon}\t")
        string += f"{beacon}Exchange:\n"
        string += f"{beacon}{exchange_str}\n"
        if verbose:  # pragma: no cover
            print(string)
        return string

    def ping(self) -> None:  # pragma: no cover
        """Ping the exchange with apikey to test its validity."""
        error_msg = f"ping() not implemented for {self.__class__.__name__}"
        raise NotImplementedError(error_msg)

    def create_client(self):  # pragma: no cover # noqa
        error_msg = f"create_client() not implemented for {self.__class__.__name__}"
        raise NotImplementedError(error_msg)

    def exchange_controller(self) -> "ExchangeController":  # pragma: no cover # noqa
        error_msg = f"exchange_controller() not implemented for {self.__class__.__name__}"
        raise NotImplementedError(error_msg)

    def get_tickers(self) -> list[str]:
        """Return a list of available tickers on the exchange."""
        return self.exchange.get_tickers()


class BinanceAccount(ExchangeAccount):
    """An exchange account on the Binance exchange."""

    public_key = models.CharField(max_length=200)
    private_key = models.CharField(max_length=200)

    class Meta:  # noqa
        unique_together = ("public_key", "private_key")

    def __str__(self) -> str:
        return "BINANCE " + super().__str__()

    def ping(self) -> None:
        """Ping the exchange with apikey to test its validity."""
        request = self.exchange_controller().get_info()
        if "error" in request:
            error_msg = f"Error pinging {self.exchange.name}. Check that your API keys are correct and have the correct permissions."
            raise ExchangeAccountError.APIPermissionError(error_msg)

    def exchange_controller(self) -> BinanceController:
        """Return the Binance's controller."""
        return BinanceController(self.public_key, self.private_key)


EXCHANGE_ACCOUNT_DICT = {
    "BINANCE": BinanceAccount,
}



from typing import TypeVar

from django.apps import apps

T = TypeVar("T", bound="FindableClass")


class FindableClass:
    """Class to find the correct subclass of a django model."""

    def find(self: T) -> T:
        """Find the correct subclass of this django model."""
        instance = self

        for subclass in self.__class__.__subclasses__():
            django_model = apps.get_model(self._meta.app_label, f"{subclass.__name__}")
            try:
                instance = django_model.objects.get(pk=instance.pk)
                break
            except django_model.DoesNotExist:
                pass
        return instance



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response

from django_napse.api.custom_permissions import HasMasterKey
from django_napse.api.custom_viewset import CustomViewSet
from django_napse.api.exchanges.serializers.exchange_account_detail_serializer import ExchangeAccountDetailSerializer
from django_napse.api.exchanges.serializers.exchange_account_serializer import ExchangeAccountSerializer
from django_napse.core.models import Exchange, ExchangeAccount
from django_napse.utils.constants import EXCHANGES

if TYPE_CHECKING:
    from django.db.models.query import Queryset
    from rest_framework import serializers
    from rest_framework.request import Request


class ExchangeAccountView(CustomViewSet):
    """Define endpoints for ExchangeAccount."""

    permission_classes: ClassVar = [HasMasterKey]
    serializer_class = ExchangeAccountSerializer

    def get_object(self) -> ExchangeAccount:
        """Get the ExchangeAccount instance for a detail endpoint."""
        return self.get_queryset().get(uuid=self.kwargs["pk"])

    def get_queryset(self) -> Queryset:
        """Get all ExchangeAccount instances."""
        return ExchangeAccount.objects.all()

    def get_serializer_class(self) -> serializers.Serializer:
        """Get the correct serializer for the action."""
        actions: dict = {
            "list": ExchangeAccountSerializer,
            "retrieve": ExchangeAccountDetailSerializer,
        }
        result = actions.get(self.action)
        return result if result else super().get_serializer_class()

    def list(self, request: Request) -> None:  # noqa: ARG002
        """List all ExchangeAccount instances."""
        serializer = self.get_serializer(self.get_queryset(), many=True)
        return Response(serializer.data, status=status.HTTP_200_OK)

    def retrieve(self, request: Request, pk: int | str | None = None) -> None:  # noqa: ARG002
        """Return the detail info of the given ExchangeAccount."""
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        return Response(serializer.data, status=status.HTTP_200_OK)

    def create(self, request: Request) -> None:
        """Create an ExchangeAccount instance."""
        if "name" not in request.data:
            return Response({"error": "Missing name"}, status=status.HTTP_400_BAD_REQUEST)
        if "exchange" not in request.data:
            return Response({"error": "Missing exchange"}, status=status.HTTP_400_BAD_REQUEST)
        if "testing" not in request.data:
            return Response({"error": "Missing testing"}, status=status.HTTP_400_BAD_REQUEST)

        # TODO (Xénépix) : Rework the following part to use a serializer. # noqa
        exchange = Exchange.objects.get(name=request.data["exchange"])
        exchange_account = ExchangeAccount.objects.create(
            exchange=exchange,
            name=request.data["name"],
            testing=request.data["testing"],
            description=request.data.get("description", ""),
        )
        serializer = self.get_serializer(exchange_account)
        return Response(serializer.data, status=status.HTTP_201_CREATED)

    def destroy(self, request: Request, pk: int | str | None = None) -> None:  # noqa: ARG002
        """Destroy an ExchangeAccount instance."""
        instance = self.get_object()
        instance.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)

    def patch(self, request: Request, pk: int | str | None = None) -> None:  # noqa: ARG002
        """Patch an ExchangeAccount instance."""
        instance = self.get_object()
        if "name" in request.data:
            instance.name = request.data["name"]
        if "description" in request.data:
            instance.description = request.data["description"]
        # TODO (Xénépix) : Rework the following part to use a serializer. # noqa

        instance.save()
        serializer = self.get_serializer(instance)
        return Response(serializer.data, status=status.HTTP_200_OK)

    @action(detail=False, methods=["GET"])
    def possible_exchanges(self, request: Request) -> None:  # noqa: ARG002
        """Return the list of possible Exchange to build the ExchangeAccount."""
        return Response(list(EXCHANGES), status=status.HTTP_200_OK)



from django.db import models

from django_napse.utils.constants import PLUGIN_CATEGORIES
from django_napse.utils.errors import PluginError


class PluginManager(models.Manager):
    def create(self, strategy):
        if self.model.plugin_category() not in PLUGIN_CATEGORIES:
            error_msg = f"Invalid plugin category: {self.model.plugin_category()}"
            raise PluginError.InvalidPluginCategory(error_msg)
        plugin = self.model(strategy=strategy, category=self.model.plugin_category())
        plugin.save()
        return plugin



from django.conf import settings
from rest_framework import status
from rest_framework.response import Response

from django_napse.api.custom_viewset import CustomViewSet
from django_napse.api.orders.serializers import OrderSerializer
from django_napse.core.models import Order


class OrderView(CustomViewSet):
    """."""

    # permission_classes = [HasAPIKey, HasSpace]
    permission_classes = []
    serializer_class = OrderSerializer

    def get_queryset(self):
        print(f"count: {Order.objects.count()}")
        return Order.objects.all()

    def list(self, request):
        """For test & debug purposes only."""
        if not settings.DEBUG:
            return Response(status=status.HTTP_404_NOT_FOUND)
        serializer = self.get_serializer(self.get_queryset(), many=True)
        return Response(serializer.data, status=status.HTTP_200_OK)



from __future__ import annotations

import operator
import uuid
from typing import ClassVar

from rest_framework.serializers import ValidationError

from django_napse.utils.serializers.fields import Field


class BaseSerializer:
    """Base class for Serializer."""

    fields_map: ClassVar[dict[str, Field]] = {}
    compiled_fields: ClassVar[
        tuple[
            str,
            callable,
            bool,
            callable,
            bool,
            bool,
        ]
    ] = {}
    validators_map: ClassVar[dict[str, tuple[bool, callable]]] = {}


class MetaSerializer(type):
    """Define the creation of Serializer classes."""

    @staticmethod
    def _compile_fields(fields: dict[str, Field], serializer_cls: Serializer) -> list[tuple]:
        """Compile a field to give an easy access to all elements of each fields."""

        def _compile_field(field: dict[str, Field], name: str, serializer_cls: Serializer) -> tuple:
            """Compile field's elements to tuple."""
            if field.source is not None and "." in field.source:
                getter: set = [operator.attrgetter(attr) for attr in field.source.split(".")]
                getter_is_generator = True
            else:
                getter = field.as_getter(name, serializer_cls) or operator.attrgetter(name)
                getter_is_generator = False
            return (
                name,
                field.default,
                field.to_value,
                getter,
                field.getter_takes_serializer,
                getter_is_generator,
            )

        return [_compile_field(field, name, serializer_cls) for name, field in fields.items()]

    def __new__(cls: MetaSerializer, name: str, bases: tuple, attrs: dict) -> Serializer:
        """Define how to build Serializer classes.

        Args:
            cls: MetaClass class.
            name: Class name.
            bases: Bases of the class.
            attrs: Method and attributs of the class (like Fields).
        """
        fields = {}

        # Get direct Fields from attributs
        fields = {key: value for key, value in attrs.items() if isinstance(value, Field)}
        for key in fields:
            del attrs[key]

        # Make serializer class
        serializer_cls = super(MetaSerializer, cls).__new__(cls, name, bases, attrs)

        # Retrieve Fields from parent classes
        for obj in serializer_cls.__mro__[::-1]:
            if issubclass(obj, BaseSerializer):
                fields.update(obj.fields_map)

        serializer_cls.fields_map = fields
        serializer_cls.compiled_fields = cls._compile_fields(fields, serializer_cls)
        serializer_cls.validators_map = {
            name: (
                field.required,
                field.validate,
            )
            for name, field in fields.items()
            if field.validate
        }

        return serializer_cls


class Serializer(BaseSerializer, Field, metaclass=MetaSerializer):  # noqa
    Model = None
    # If read_only is True, the serializer will not be able to make interaction with the 'Model' class.
    # But it can still be used to serialize or validate data.
    read_only = False

    def __init__(self, instance=None, data=None, many=False, **kwargs):  # noqa
        self._instance = instance
        self._data = data
        self._many = many
        self._validated_data: dict[str, any] | None = None

        if self._data is not None:
            self.validate_data(self._data)

        # To use serializer in serializer
        super().__init__(**kwargs)

        # Save additionnal kwargs for compatibility
        self._kwargs = kwargs

    @property
    def data(self):  # noqa
        if self._data is not None:
            return self._data
        return self.to_value(instance=self._instance)

    @property
    def validated_data(self):  # noqa
        if self._validated_data is None:
            error_msg: str = "Data are invalid"
            raise ValueError(error_msg)
        return self._validated_data

    def to_value(self, instance: object | list[object] | None = None) -> any:
        """Serialize instance."""
        # TODO: make a try & except TypeError | AttributeError (log + return 400) # noqa
        instance = instance or self._instance

        if instance is None:
            return None

        if self._many:
            try:
                len(instance)
            except TypeError:
                instance = instance.all()
            return [self._serialize(ist, self.compiled_fields) for ist in instance]
        return self._serialize(instance, self.compiled_fields)

    def _serialize(self, instance: object, fields: dict[str, Field]) -> dict[str, any]:
        def _get_value(instance, getter, getter_takes_serializer, getter_is_generator):  # noqa
            if getter_is_generator:
                value = instance
                for get in getter:
                    value = get(value)
            else:
                value = getter(self, instance) if getter_takes_serializer else getter(instance)
            return value

        serialized_instance = {}
        for name, default, to_value, getter, getter_takes_serializer, getter_is_generator in fields:
            # Default wrapper around getter for optimization
            if default is None:
                value = _get_value(
                    instance,
                    getter,
                    getter_takes_serializer,
                    getter_is_generator,
                )
            else:
                try:
                    value = _get_value(
                        instance,
                        getter,
                        getter_takes_serializer,
                        getter_is_generator,
                    )
                except AttributeError:
                    value = default

            if to_value:
                value = to_value(value)
            serialized_instance[name] = value
        return serialized_instance

    def validate_data(self, data: dict) -> any:
        """Used for validation with serializer.

        Please to not use this method.
        """
        if not isinstance(data, dict):
            error_msg: str = "Data must be a dictionary."
            raise ValidationError(error_msg)
        validated_data = {}
        for name, (required, validator) in self.validators_map.items():
            elt = data.get(name)
            if elt is None:
                if required:
                    error_msg: str = f"Field {name} is required."
                    raise ValidationError(error_msg)
                continue

            result = validator(elt)
            if not result:
                error_msg: str = f"Field {name} is invalid."
                raise ValidationError(error_msg)
            if isinstance(result, bool):
                validated_data[name] = elt
            else:
                validated_data[name] = result

        self._validated_data = validated_data
        return validated_data

    def validate(self, data: dict | str | int | uuid.UUID) -> any:
        """Used for automatic validation with serializer.

        Please to not use this method.
        """
        if self.read_only:
            error_msg: str = "This serializer is read only."
            raise ValueError(error_msg)

        if isinstance(data, (str, int, uuid.UUID)):
            return self.get(uuid_or_id=data)

        return self.validate_data(data)

    def _model_checks(self, validated_data):  # noqa
        if self.read_only:
            error_msg: str = "This serializer is read only."
            raise ValueError(error_msg)

        if not self.Model:
            error_msg: str = "Model is not defined."
            raise ValueError(error_msg)

        if not validated_data:
            error_msg: str = "Data are not validated."
            raise ValueError(error_msg)

    def get(self, uuid_or_id=None):  # noqa
        """Retrieve instance from data."""
        try:
            instance = self.Model.objects.get(uuid=uuid_or_id) if hasattr(self.Model, "uuid") else self.Model.objects.get(id=uuid_or_id)
        except (self.Model.DoesNotExist, TypeError):
            error_msg: str = f"Instance with id {uuid_or_id} does not exist."
            raise ValidationError(error_msg) from None
        return instance

    def create(self, validated_data=None):  # noqa
        validated_data = validated_data or self._validated_data
        self._model_checks(validated_data=validated_data)
        return self.Model.objects.create(**validated_data)

    def update(self, instance, validated_data=None):  # noqa
        validated_data = validated_data or self._validated_data
        self._model_checks(validated_data=validated_data)
        for key, value in validated_data.items():
            setattr(instance, key, value)
        instance.save()
        return instance



from .architecture import *
from .architectures import *
from .bot import *
from .config import *
from .controller import *
from .implementations import *
from .plugin import *
from .plugins import *
from .strategy import *



class WalletError:
    """Base class for wallet errors."""

    class SpendError(Exception):
        """Raised when a wallet cannot spend."""

    class TopUpError(Exception):
        """Raised when a wallet cannot top up."""

    class CreateError(Exception):
        """Raised when a wallet cannot be created."""

    class UnavailableTicker(Exception):
        """Raised when a ticker is the wallet's currencies."""



from django_napse.core.models import History, HistoryDataPoint, HistoryDataPointField
from django_napse.utils.constants import HISTORY_DATAPOINT_FIELDS
from django_napse.utils.errors import HistoryError
from django_napse.utils.model_test_case import ModelTestCase

"""
python tests/test_app/manage.py test tests.django_tests.db.histories -v2 --keepdb --parallel
"""


class HistoryTestCase:
    model = History

    def simple_create(self):
        return self.model.objects.create()

    def test_data_points(self):
        info = [
            [
                {"key": HISTORY_DATAPOINT_FIELDS.AMOUNT, "value": 1},
                {"key": HISTORY_DATAPOINT_FIELDS.ASSET, "value": "BTC"},
                {"key": HISTORY_DATAPOINT_FIELDS.PRICE, "value": 123.0},
                {"key": HISTORY_DATAPOINT_FIELDS.MBP, "value": 100},
                {"key": HISTORY_DATAPOINT_FIELDS.LBO, "value": 7},
            ]
            for _ in range(100)
        ]
        history = self.simple_create()
        for data_point_info in info:
            data_point = HistoryDataPoint.objects.create(history=history)
            for field_info in data_point_info:
                HistoryDataPointField.objects.create(history_data_point=data_point, **field_info)
        self.assertEqual(history.data_points.count(), len(info))
        target_types = {
            HISTORY_DATAPOINT_FIELDS.AMOUNT: "int",
            HISTORY_DATAPOINT_FIELDS.ASSET: "str",
            HISTORY_DATAPOINT_FIELDS.PRICE: "float",
            HISTORY_DATAPOINT_FIELDS.MBP: "int",
            HISTORY_DATAPOINT_FIELDS.LBO: "int",
        }
        for data_point in history.data_points.all():
            for field in data_point.fields.all():
                self.assertEqual(field.target_type, target_types[field.key])

    def test_invalid_data_point(self):
        history = self.simple_create()
        data_point = HistoryDataPoint.objects.create(history=history)

        with self.assertRaises(HistoryError.InvalidDataPointFieldKey):
            HistoryDataPointField.objects.create(history_data_point=data_point, key="INVALID", value=1)


class HistoryBINANCETestCase(HistoryTestCase, ModelTestCase):
    exchange = "BINANCE"



class NapseError:
    """Base class for exceptions in this module."""

    class SettingsError(Exception):
        """Exception raised for errors in the settings file."""

    class RandomGenrationError(Exception):
        """Exception raised for errors in random generation processes."""



from datetime import datetime

from django.test import TestCase
from pytz import UTC

from django_napse.core.models import Bot, Strategy
from django_napse.simulations.models import SimulationQueue

"""
python tests/test_app/manage.py test tests.django_tests.bots.test_bot -v2 --keepdb --parallel
"""


class StrategyDefaultTestCase:
    def simple_create(self):
        return self.model.objects.create(config=self.config, architecture=self.architecture)

    @property
    def config(self):
        return self.model.config_class().objects.create(space=self.space, settings=self.config_settings)

    @property
    def architecture(self):
        return self.model.architecture_class().objects.create(constants=self.architecture_constants)

    def test_simulation(self):
        bot = Bot.objects.create(name="Test Bot", strategy=self.simple_create())
        simulation_queue = SimulationQueue.objects.create(
            space=self.space,
            bot=bot,
            start_date=datetime(2022, 1, 1, tzinfo=UTC),
            end_date=datetime(2022, 1, 5, tzinfo=UTC),
            investments={"USDT": 10000},
        )

        simulation_queue.run_quick_simulation(verbose=False)


class BotTypeCkeck(TestCase):
    def test_bot_type(self):
        subclasses = []
        for subclass_level in StrategyDefaultTestCase.__subclasses__():
            subclasses += subclass_level.__subclasses__()
        tested_strategies = {*[subclass.model for subclass in subclasses]}
        strategies = {*Strategy.__subclasses__()}
        if tested_strategies != strategies:
            error_msg = "You have untested Strategies. Check out the documentation to see how to test them (spoiler, it's really easy!)."
            error_msg += str(tested_strategies) + str(strategies)
            raise ValueError(error_msg)



from django.db import models


class SimulationInvestedCurrency(models.Model):
    owner = models.ForeignKey("Simulation", on_delete=models.CASCADE, related_name="investments")
    ticker = models.CharField(max_length=10)
    amount = models.FloatField(default=0)

    class Meta:
        unique_together = ("owner", "ticker")

    def __str__(self):
        return f"SIMULATION INVESTED CURRENCY {self.pk}"

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}Currency ({self.pk=}):\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.owner=}\n"
        string += f"{beacon}\t{self.mbp=}\n"
        string += f"{beacon}\t{self.ticker=}\n"
        string += f"{beacon}\t{self.amount=}\n"

        if verbose:  # pragma: no cover
            print(string)
        return string

    @property
    def testing(self):
        return self.wallet.testing


class SimulationQueueInvestedCurrency(models.Model):
    owner = models.ForeignKey("SimulationQueue", on_delete=models.CASCADE, related_name="investments")
    ticker = models.CharField(max_length=10)
    amount = models.FloatField(default=0)

    class Meta:
        unique_together = ("owner", "ticker")

    def __str__(self):
        return f"SIMULATION QUEUE INVESTED CURRENCY {self.pk}"

    def info(self, verbose=True, beacon=""):
        string = ""
        string += f"{beacon}Currency ({self.pk=}):\n"
        string += f"{beacon}Args:\n"
        string += f"{beacon}\t{self.owner=}\n"
        string += f"{beacon}\t{self.mbp=}\n"
        string += f"{beacon}\t{self.ticker=}\n"
        string += f"{beacon}\t{self.amount=}\n"

        if verbose:  # pragma: no cover
            print(string)
        return string

    @property
    def testing(self):
        return self.wallet.testing



from django.db import models

from django_napse.utils.constants import MODIFICATION_STATUS

from .modification import Modification


class ConnectionModification(Modification):
    connection_specific_arg = models.ForeignKey("ConnectionSpecificArgs", on_delete=models.CASCADE, related_name="modifications")

    def apply(self):
        connection_specific_arg, self = self.apply__no_db()
        connection_specific_arg.save()
        self.save()

    def apply__no_db(self, **kwargs):
        self.connection_specific_arg.value = self.get_value(current_value=self.connection_specific_arg.get_value(), **kwargs)
        self.status = MODIFICATION_STATUS.APPLIED
        return self.connection_specific_arg, self
